% [xmin, fx] = brentWrap( funk, iX, oDir, ix )
%
% Synopsis
%     This is a wrapper for brent's 1D method for minimising a function, which
%     uses parabolic interpolation to track down the point where it can,
%     otherwise it reverts to a golden search routine.

%Example of use
%  %Setup
%  fn( in.tmp, in.img, 'std', ttPar );
%  %Bracket and minimise
%  fStep = 0.01;
%  [fxa, fxb, fxc, fya, fyb, fyc] = mnbrak( @fn, 0, 1, oX, oDir(:,mi) );
%  [ftX oY] = brent( @fn, fxa, fxb, fxc, oX, oDir(:,mi), fStep ); %ft/10;

function [xmin, fx] = brentWrap( funk, iX, oDir, ix )
% BRENTWRAP - A wrapper function for brent
%   
  if ~exist('ix','var')
    ix = [0 1 0.01];
  end
  [fxa, fxb, fxc, fya, fyb, fyc] = mnbrak( funk, ix(1), ix(2), iX, oDir );
  [xmin, fx] = brent( funk, fxa, fxb, fxc, iX, oDir, ix(3) ); %ft/10;
  
%===========================================================================
% This is brent's 1D method for minimising a function, which uses parabolic
% interpolation to track down the point where it can, otherwise it reverts
% to a golden search routine.
function [xmin, fx] = brent( funk, ax, bx, cx, X0, A, iMinStep )

  global gEvals;
  
  itermx = 100;
  CGOLD = 0.3819660;
  ZEPS = 1e-10;
  pFTolY = 0.01; %Fractional tolerance for Y

  d=0; 
  e=0; %Distance moved on 2nd last step
  a=min([ax cx]); %a and b are in ascending order
  b=max([ax cx]);

  x=bx;w=bx;v=bx; %Initialise
  fx=funk( A*x+X0 ); gEvals = gEvals+1;
  fw=fx; fv=fx;

  %Main loop
  for iter=1:itermx
    xm = 0.5*(a+b);
    tol1=pFTolY*abs(x)+ZEPS; 
    tol2=2*tol1;
    
    %Test for done
    if(abs(x-xm) <= (tol2-0.5*(b-a))) break; end
    if exist('fu') if (fu==fv) break; end; end
    %if exist('u') if u-v<iMinStep/2 break; end; end
    
    %Construct trial parabolic fit
    if (abs(e) > tol1)
      r=(x-w)*(fx-fv);
      q=(x-v)*(fx-fw);
      p=(x-v)*q-(x-w)*r;
      q=2*(q-r);
      if (q>0) p=-p; end;
      q=abs(q);
      etemp=e;
      e=d;
      %If parabolic fit is not acceptable, use take golden section
      if( abs(p) >= abs(0.5*q*etemp) || p<=q*(a-x) || p >= q*(b-x) )
        if (x>=xm) e=a-x; else e=b-x; end;
        d = CGOLD*e;
      else %Otherwise Take parabolic step
        d=p/q;
        u=x+d;
        if (u-a < tol2 || b-u < tol2)
          d = tol1*sign( xm-x );
        end
      end
    else
      if (x>=xm) e=a-x; else e=b-x; end;
      d = CGOLD*e;
    end
    if abs(d) >= tol1
      u=x+d;
    else
      u=x+tol1*sign(d);
    end
    %This is the one function evaluation per iteration
    fu = funk( A*u+X0 ); gEvals = gEvals+1;
    if fu <= fx
      if (u >= x) a=x; else b=x; end
      v=w; w=x; x=u;
      fv=fw; fw=fx; fx=fu;
    else
      if (u<x) a=u; else b=u; end
      if (fu<=fw || w==x)
        v=w; w=u;
        fv=fw; fw=fu;
      elseif (fu<=fv || v==x || v==w)
        v=u;
        fv=fu;
      end
    end
  end
  xmin = x;

%=============================================================================
% This function brackets a minimum on a 1D line with three points ax ab and
% cx. Their respective values are given as y. The minimum is actually being
% sought on a line through a multi-D space which is defined as y = f(A*x+X0)
% The function searches in a downhill direction taking ever larger steps
function [ax, bx, cx, fa, fb, fc] = mnbrak( funk, ax, bx, X0, A )
  
  global gEvals;
  GOLD = 1.618034; %Default ratio to magnify steps
  GLIMIT = 5*GOLD; %Maximum magnification fo a parabolic fit step
  TINY = 1e-20;

  fa = funk( A*ax+X0 ); gEvals = gEvals+1;
  fb = funk( A*bx+X0 ); gEvals = gEvals+1;
  t = sortrows( [fa ax; fb bx], 1 ); 
  fa = t(2,1); ax = t(2,2);
  fb = t(1,1); bx = t(1,2);
  %First guess for cx
  cx = bx + GOLD*(bx-ax);
  fc = funk( A*cx+X0 ); gEvals = gEvals+1;
  %Main loop
  while fb > fc
    r = (bx-ax)*(fb-fc); %Keep returning here until we bracket
    q = (bx-cx)*(fc-fa); %Get u you by parabolic extrapolation from a,b,c
    u=bx-((bx-cx)*q-(bx-ax)*r)/2*max([abs(q-r) TINY]);
    if (q-r<0) u=-u; end
    ulim = bx+GLIMIT*(cx-bx); %No further than this is allowed
    if ((bx-u)*(u-cx) > 0) %Parabolic u is betw b,c - try it! (nice trick!)
      fu = funk( A*u+X0 ); gEvals = gEvals+1;

      if(fu<fc) %Got minimu between b and c
        ax=bx; bx = u; fa=fb; fb=fu;
        return;nntol
      elseif (fu>fb) %Got minimum between a and u
        cx=u; fc=fu;
        return;
      end
      u=cx+GOLD*(cx-bx); %Default fit no use, Magnify
      fu = funk( A*u+X0 ); gEvals = gEvals+1;
    elseif ((cx-u)*(u-ulim) > 0) %Parabolic fit between c and its limit
      fu = funk( A*u+X0 ); gEvals = gEvals+1;
      if (fu< fc)
        bx=cx; cx=u; u=cx+GOLD*(cx-bx);
        fb=fc; fc=fu; fu = funk( A*u+X0 ); gEvals = gEvals+1;
      end
    elseif ((u-ulim)*(ulim-cx) >= 0) %Limit u to maximum allowed value
      u=ulim;
      fu = funk( A*u+X0 ); gEvals = gEvals+1;
    else %Reject parabolic u and magnify
      u=cx+GOLD*(cx-bx);
      fu = funk( A*u+X0 ); gEvals = gEvals+1;
    end
    ax=bx; bx=cx; cx=u; %Eliminate oldest point and continue
    fa=fb; fb=fc; fc=fu;
  end
