/*****************************************************************
 * cimgmatlabplugin.h
 * ------------------
 *
 * cimgmatlabplugin is a "plugin" for the CImg library that allows 
 * to convert CImg<T> images from/to matlab arrays, so that CImg
 * can be used to write matlab mex files.
 * How to use it:
 * file cimgmatlabplugin.h must be in a directory that the 
 * compiler can locate.
 * Prior to include CImg.h, mex.h must be included first, else it
 * will result in a compiler error.
 * after the inclusion of mex.h, one must define the macro
 * cimg_plugin as "cimgmatlabplugin.h" or <cimgmatlabplugin.h>
 *
 * // my fantastic mex file source code
 * #include <mex.h>
 * ...
 * #define cimg_plugin  "cimgmatlabplugin.h"
 * #include <CImg.h>
 * ...
 *
 *
 * Copyright: Francois Lauze, The IT University of Copenhagen.
 * Licence: the Gnu Lesser General Public License
 *          http://www.gnu.org/copyleft/lesser.html
 * or any compatible one, as long as there are no problems with 
 * the CeCill license used by CImg ...
 *****************************************************************/

//Warning: several changes have been made, to make this fit the matlab
//conventions properly again. Look for //Nick

#if !defined mex_h
#error the file mex.h must be included!
#endif

#if !defined cimg_version
#error cimgmatlabplugin.h requires that CImg.h is included
#endif




// begin of included methods
// They are just added as member functions / constructor
// for the CImg<T> class.

private:
// internally used to transfer Matlab array values to CImg<> objects 
// check wether the array type is a "numerical" one.
static int isNumericalClassID(mxClassID id) {
  switch (id) {
    // all these constants are defined in mex.h or files included from it
  case mxLOGICAL_CLASS:
  case mxDOUBLE_CLASS:
  case mxSINGLE_CLASS:
  case mxINT8_CLASS:
  case mxUINT8_CLASS:
  case mxINT16_CLASS:
  case mxUINT16_CLASS:
  case mxINT32_CLASS:
  case mxUINT32_CLASS:
    return 1;
  default:
    return 0;
  }
}


// driving routine that will copy the content of
// a MatLab array to this->data
void makeImageFromMatlabData(const mxArray *matlabArray, mxClassID id) {
  if (id == mxLOGICAL_CLASS) {
    mxLogical *mdata = mxGetLogicals(matlabArray);
    cpMatlabData((const mxLogical *)mdata);
  } else {
    void *mdata = mxGetPr(matlabArray);
    switch (id) {
    case mxDOUBLE_CLASS:
      cpMatlabData((const double *)mdata);
      break;
    case mxSINGLE_CLASS:
      cpMatlabData((const float *)mdata);
      break;
    case mxINT8_CLASS:
      cpMatlabData((const signed char *)mdata);
      break;
    case mxUINT8_CLASS:
      cpMatlabData((const unsigned char *)mdata);
      break;
    case mxINT16_CLASS:
      cpMatlabData((const short *)mdata);
      break;
    case mxUINT16_CLASS:
      cpMatlabData((const unsigned short *)mdata);
      break;
    case mxINT32_CLASS:
      cpMatlabData((const int *)mdata);
      break;
    case mxUINT32_CLASS:
      cpMatlabData((const unsigned int *)mdata);
      break;
    }
  }
}
    
// the actual memory copy and base type conversion
// (through C++ genericity mechanisms) is then
// performed by this routine that handles the annoying 
// x-y problem of MatLab: we switch line and column
// storage: Matlab A(x,y) becones CImg img(y,x)
template <typename t> void cpMatlabData(const t* mdata) 
{
  //cimg_mapXYZV(*this, x, y, z, v) 
  //  {
  //    (*this)(x, y, z, v) = (T)(mdata[((v*depth + z)*width+x)*height+y]);
  //  }
  for( unsigned int i=0; i<size(); i++ ) data[i] = (T) mdata[i]; //Nick
}

public:

// Construct a CImg<T> object from a MatLab mxArray.
// The Matlab array must be AT MOST 4-dimensional.
// the boolean argument vdata is employed in the 
// case the the input mxArray has dimension 3,
// say M x N x K.
// In that case, if vdata is true, the last dimension
// is assumed to be "vectorial" and the resulting CImg<T>
// object has dimension N x M x 1 x K. Otherwise, the
// resulting object has dimension N x M x K x 1.
// When Matlab array has dimension 2 or 4, vdata
// has no effects.
CImg(const mxArray *mArray, bool vdata = false) 
{
  int nbdims = mxGetNumberOfDimensions(mArray);
  mxClassID classID = mxGetClassID(mArray);
  if (nbdims > 4 || !isNumericalClassID(classID)) 
    {
      data=NULL; width=height=depth=dim=0;
#if cimg_debug>1
      cimg::warn(1,"Matlab array is more than 4D"
		 " or/and not numerical, returning null image.");
#endif	    
    } 
  else 
    {
      const int *dims = mxGetDimensions(mArray);
      depth = dim = 1;
      //width = dims[1]; 
      //height = dims[0];
      width = dims[0];  //Nick
      height = dims[1]; //Nick
      if (nbdims == 4) 
	{
	  depth = dims[2];  
	  dim = dims[3];
	} 
      else if (nbdims == 3) 
	{
	  if (vdata) dim = dims[2];
	  else depth = dims[2];
	}
      data = new T[size()];
      makeImageFromMatlabData(mArray, classID);
    }
}


	
//! 0perator =. Copy  mxMarray data \p mArray into the current image
// Works as the previous constructor, but without the vdata stuff. 
CImg & operator=(const mxArray *mArray) {
  int nbdims = mxGetNumberOfDimensions(mArray);
  int classID = mxGetClassID(mArray);
  if (nbdims > 4 || !isNumericalClassID(classID)) {
    delete [] data; data = NULL; width=height=depth=dim=0;
#if cimg_debug>1
    cimg::warn(1,"Matlab array is more than 4D "
	       "or/and not numerical, returning null image.");
#endif	    
  } else {
    const int *dims = mxGetDimensions(mArray);
    depth = dim = 1;
    //width = dims[1]; 
    //height = dims[0];
    width = dims[0];  //Nick
    height = dims[1]; //Nick

    if (nbdims > 2) depth = dims[2];
    if (nbdims > 3) dim = dims[3];
    delete [] data;
    data = new T[size()];
    makeImageFromMatlabData(mArray, classID);
  }
}

private:      
// private routines used for transfering a CImg<T> to a mxArray
// here also, we have to exchange the 
template <typename c>
void populate_maltlab_array(c *mdata) const {
  //  cimg_mapXYZV(*this, x, y, z, v) 
  //    mdata[((v*depth + z)*width+x)*height+y] = (c)(*this)(x, y, z, v);
  for( unsigned int i=0; i<size(); i++ ) mdata[i] = (c) data[i]; //Nick
}
// the specialized version for "logical" entries
void populate_maltlab_array(mxLogical *mdata) const {
  //  cimg_mapXYZV(*this, x, y, z, v) 
  //    mdata[((v*depth + z)*width+x)*height+y] = (*this)(x, y, z, v)!=0;
  for( unsigned int i=0; i<size(); i++ ) mdata[i] = data[i]!=0; //Nick
}
    
    
public:
//! export a CImg image to a matlab array.  
mxArray *toMatlab(mxClassID id = mxDOUBLE_CLASS, bool squeeze = false) const {
  if (!isNumericalClassID(id)) {
#if cimg_debug>1
    cimg::warn(1,"Invalid Matlab Class Id Specified.");
#endif
    return NULL;
  }
  int dims[4];
  //dims[0] = height;
  //dims[1] = width;
  dims[0] = width; //Nick
  dims[1] = height; //Nick

  dims[2] = depth;
  dims[3] = dim;
  if (squeeze && depth == 1) {
    dims[2] = dim;
    dims[3] = 1;
  }
  mxArray *mArray = mxCreateNumericArray(4,dims,id,mxREAL);
  if (id == mxLOGICAL_CLASS) {
    mxLogical *mdata = mxGetLogicals(mArray);
    populate_maltlab_array(mdata);
  } else {
    void *mdata = mxGetPr(mArray);
    switch (id) {
    case mxDOUBLE_CLASS:
      populate_maltlab_array(( double *)mdata);
      break;
    case mxSINGLE_CLASS:
      populate_maltlab_array(( float *)mdata);
      break;
    case mxINT8_CLASS:
      populate_maltlab_array(( signed char *)mdata);
      break;
    case mxUINT8_CLASS:
      populate_maltlab_array(( unsigned char *)mdata);
      break;
    case mxINT16_CLASS:
      populate_maltlab_array(( short *)mdata);
      break;
    case mxUINT16_CLASS:
      populate_maltlab_array(( unsigned short *)mdata);
      break;	
    case mxINT32_CLASS:
      populate_maltlab_array(( int *)mdata);
      break;
    case mxUINT32_CLASS:
      populate_maltlab_array(( unsigned int *)mdata);
      break;
    }	
  }
  return mArray;
}

// end of included methods
