function [F G H H2 H3] = mibwds(T,R,W,PZ)
% MIBWDS - Gradient and Hessian for Mutual Information
%   Synopsis:
%     Given a template (T) and warping it to a particular position on a
%     reference image (R) for warp parameters (W), computes the Mutual
%     Information between the shared regions, along with a Jacobian and
%     Hessian.
%   Inputs:
%     T - template image
%     R - reference image
%     W - warp parameters
%   Outputs:
%     F - Function Value
%     G - Gradient
%     H - Hessian
  
  G = []; H = []; H2 = []; H3 = [];
  W(5:6)=W(5:6)+1; %Fix up coordinates
  
  if ~exist('PZ','var'), PZ=1; end
  PZ1 = sum(PZ,2);
  
  Nt = floor(max(T(:)))+1;
  Nr = floor(max(R(:)))+1;
  
  X = coordVals(T); %Generate coordinate matrix  
  rot = warp2RotMatrix(W); %Generate Rotation Matrix
  Xw = rot*X; %Get warped coordinates
  
  Rw = getRegion(R,Xw,size(T)); %Get the warped region

  fH = conv2( PZ, jointHist(T,Rw,[Nt Nr]) ); %Get the joint histogram

  %Generate MI
  tValid = find( fH~= 0);
  tH = fH .* repmat(sum(fH,1),Nt,1).^-1 .* repmat(sum(fH,2),1,Nr).^-1;
  F = sum( fH(tValid).*log( tH(tValid) ) );

  if nargout < 2, return; end  
  %-------------------------------------------------------------------
  %Generate Jacobian histogram
  %[Tdx Tdy] = gradient(T); %Get gradient image of T
  Tdx = [diff(T,1,1); zeros(1,size(T,2))];
  Tdy = [diff(T,1,2)  zeros(size(T,1),1)];
  
  dWdv = getJacobian(T); %Generate warp derivative
  Tsd = imgMul({Tdx,Tdy},dWdv); %Get steepest descent image

  for c1=1:6, 
    fHJ{c1} = conv2( conv2([-1 1]',PZ), jointHist(T,Rw,[Nt Nr],Tsd{c1}) );
  end
  
  %Generate the MI jacobian
  fHu = conv2( [1 1]'/2, fH );
  tH  = exp(1) * fHu .* repmat(sum(fHu,1),Nt+1,1).^-1 ...
        .* repmat(sum(fHu,2),1,Nr).^-1;
  tValid = find(fHu~=0);
  for c1 = 1:6
    G(c1) = sum( log(tH(tValid)) .* fHJ{c1}(tValid) );
  end
  
  if nargout < 3, return; end;
  %-------------------------------------------------------------------  
  %Generate the MI hessian - This may be pre-computed
  tDelTdwdvSqd = imgMul(Tsd',Tsd); %  (Del T dw/dv)^2
  %[Tdxx Tdxy] = gradient(Tdx);  %  Del^2 T / (dw/dv)^2
  %[Tdyx Tdyy] = gradient(Tdy);
  Tdxx = [diff(Tdx,1,1); zeros(1,size(T,2))];
  Tdxy = [diff(Tdx,1,2)  zeros(size(T,1),1)];
  Tdyx = [diff(Tdy,1,1); zeros(1,size(T,2))];
  Tdyy = [diff(Tdy,1,2)  zeros(size(T,1),1)];
  
  tDel2T_dwdvSqd = imgMul( imgMul(dWdv',{Tdxx Tdxy; Tdyx Tdyy}), dWdv );
  
  for c1=1:6, for c2=1:6,
    ptddf1{c1,c2} = conv( conv([-1 2 -1]',PZ1),...
                          singleHist(T,Nt,tDelTdwdvSqd{c1,c2}) );
    ptddf2{c1,c2} = conv( conv([-.5 0 .5]',PZ1),...
                          singleHist(T,Nt,tDel2T_dwdvSqd{c1,c2}) );
  end; end
  
  pt = conv([1 2 1]'/4, singleHist(T,Nt));
  for c1=1:6
    ptd{c1} = conv( conv([-.5 0 .5]',PZ1), singleHist(T,Nt,Tsd{c1}));
  end
  ptdSqd = imgMul( ptd', ptd );
  
  for c1=1:6, for c2=1:6,
      fHH{c1,c2} = -ptdSqd{c1,c2}.*pt.^-1 - ptddf1{c1,c2} - ptddf2{c1,c2};
  end; end
  
  %Assemble main hessian term
  H = zeros(6);
  tValid = find( pt~=0 );
  for c1=1:36, H(c1) = sum(fHH{c1}(tValid)); end
  
  if nargout < 4, return; end;
  %----------------------------------------------------------------
  %In-process term
  fHJsqd = imgMul(fHJ',fHJ);
  tHu = conv2([1 1]'/2,fH);
  H2 = zeros(6);
  tValid = find( tHu~=0 );
  for c1=1:36, H2(c1) = sum(fHJsqd{c1}(tValid) .* tHu(tValid).^-1); end

  if nargout < 5, return; end;
  %----------------------------------------------------------------
  %2nd order part
  for c1=1:36
    prtddf1{c1}= conv2( conv2([-1 2 -1]',PZ),...
                        jointHist(T,Rw,[Nt Nr],tDelTdwdvSqd{c1}) );
    prtddf2{c1}= conv2( conv2([-.5 0 .5]',PZ),...
                        jointHist(T,Rw,[Nt Nr],tDel2T_dwdvSqd{c1}) );
  end
  
  fHu = conv2( [1 2 1]'/4, fH );
  tH = exp(1) * fHu .* repmat(sum(fHu,1),Nt+2,1).^-1 .* ...
       repmat(sum(fHu,2),1,Nr).^-1;
  tValid = find(fHu~=0);
  H3 = zeros(6);
  for c1=1:36,
    H3(c1)= sum( log(tH(tValid)).*(prtddf1{c1}(tValid)+prtddf2{c1}(tValid) ) );
  end
  
  %Invert the hessian and get parameter update
  %warpParamDelta1 = invertWarp( - miHess \ miJac' );
  %rot = rot*warp2RotMatrix( warpParamDelta1 );
