function [F,G,H,H2,H3] = mifwds(T,R,W,PZ)
% MIFWDS - Gradient and Hessian for Mutual Information
%   Synopsis:
%     Given a template (T) and warping it to a particular position on a
%     reference image (R) for warp parameters (W), computes the Mutual
%     Information between the shared regions, along with a Jacobian and
%     Hessian.
%   Inputs:
%     T - template image
%     R - reference image
%     W - warp parameters
%   Outputs:
%     F - Function Value
%     G - Gradient
%     H - Hessian
  
  G = []; H = []; H2 = []; H3 = [];
  
  if ~exist('PZ','var'), PZ=1; end
  
  W(5:6)=W(5:6)+1; %Fix up coordinates
  
  Nt = floor(max(T(:)))+1+length(PZ)-1;
  Nr = floor(max(R(:)))+1+length(PZ)-1;
  
  X = coordVals(T); %Generate coordinate matrix  
  rot = warp2RotMatrix(W); %Generate Rotation Matrix
  Xw = rot*X; %Get warped coordinates
  
  Rw = getRegion(R,Xw,size(T)); %Get the warped region
    
  %Note convention, rows are t and cols are r
  fH = conv2(jointHist(T,Rw,[Nt Nr]),PZ); %Get the joint histogram
                              
  %Generate MI
  tValid = find( fH~= 0);
  tH = fH .* repmat(sum(fH,1),Nt,1).^-1 .* repmat(sum(fH,2),1,Nr).^-1;
  F = sum( fH(tValid).*log( tH(tValid) ) );

  if nargout < 2, return; end  
  %-------------------------------------------------------------------
  %Generate Jacobian histogram
  %[Rdx Rdy] = gradient(R); %Get the gradient image of warped Ref
  Rdx = diff(R,1,1); Rdy = diff(R,1,2);
  Rdxw = getRegion(Rdx,Xw,size(T)); %And the regions within it
  Rdyw = getRegion(Rdy,Xw,size(T));
  
  dWdv = getJacobian(T); %Generate warp derivative
  Rsd = imgMul({Rdxw,Rdyw},dWdv); %Get steepest descent image

  for c1=1:6, 
    fHJ{c1} = conv2( conv2([-1 1],PZ), jointHist(T,Rw,[Nt Nr],Rsd{c1}) );
  end

  %Generate the MI jacobian
  fHu = conv2( [1 1]/2, fH );
  tH  = exp(1) * fHu .* repmat(sum(fHu,1),Nt,1).^-1;
  tValid = find(fHu~=0);
  for c1 = 1:6,
    G(c1) = sum( log(tH(tValid)) .* fHJ{c1}(tValid) );
  end
  
  if nargout < 3, return; end;
  %-------------------------------------------------------------------  
  %Generate the MI hessian
  tH = fHu.^-1 - repmat(sum(fHu,1),Nt,1).^-1;
  for c1=1:6, for c2=1:6,
      H(c1,c2) = sum( tH(tValid) .* fHJ{c1}(tValid) .* fHJ{c2}(tValid) );
  end, end
  
  if nargout < 4, return; end;
  %-------------------------------------------------------------------  
  %2nd order part of function
  tDelRdwdvSqd = imgMul(Rsd',Rsd); %  (Del R dw/dv)^2
  
  %  Del^2 R / (dw/dv)^2
  Rdxx = diff(Rdx,1,1); Rdxy = diff(Rdx,1,2);
  Rdyx = diff(Rdy,1,1); Rdyy = diff(Rdy,1,2);
  Rdxxw = getRegion(Rdxx,Xw,size(T));  Rdxyw = getRegion(Rdxy,Xw,size(T));
  Rdyxw = getRegion(Rdyx,Xw,size(T));  Rdyyw = getRegion(Rdyy,Xw,size(T));
  
  tDel2R_dwdvSqd = imgMul( imgMul(dWdv',{Rdxxw Rdxyw; Rdyxw Rdyyw}), dWdv );
  
  for c1=1:36, 
    prtddf1{c1}= conv2( conv2([-1 2 -1],PZ), ...
                        jointHist(T,Rw,[Nt Nr],tDelRdwdvSqd{c1}) );
    prtddf2{c1}= conv2( conv2([-.5 0 .5],PZ), ...
                        jointHist(T,Rw,[Nt Nr],tDel2R_dwdvSqd{c1}) );
  end
  
  fHu = conv2( [1 2 1]/4, fH );
  tH = exp(1) * fHu .* repmat(sum(fHu,1),Nt,1).^-1;
  tValid = find(fHu~=0);
  H2 = zeros(6);
  for c1=1:36,
    H2(c1) = sum( log(tH(tValid)) .* ...
                       (prtddf1{c1}(tValid)+prtddf2{c1}(tValid) ) );
  end

  %-------------------------------------------------------------------  
  %Invert the hessian and get parameter update
  %warpParamDelta = - H \ G';
  