function [X,F,History,N] = lmmin(T,I,X,msrType,verb)
% LMMIN - Levenberg-Marquardt Minimisation for Registraion
%   Find the optimal warp parameter X that T matches the corresponding
%   warped position on I the best.
%   
  
%Initialisation
  if ~exist('msrType','var'), msrType = {'mifwds',1}; end;
  if ~exist('verb','var'), verb=[]; end;

  X=X(:)';
  History.x = []; History.y = [];  
  l = 0.001;    
  
  DxLimit = [.05 .05 .05 .05 4 4];
  
  [F G H] = fn(T,I,X,msrType); %1st function eval
  N = [0 1];
  plotCvg(X,[],F,[],verb,'begin'); %Plot progress
  
  %Outer Loop ========================================
  for c1 = 1:10,
    l = 0.001; %max( 0.1*l, 0.001 );
    hLM = H .* (eye(length(X)) * l + 1); %Get LM Hessian
    if abs(rcond(hLM))<1e-3, hLM=eye(6); end; %Check if hLM is invertible
    
    DeltaX = getDelta(G,hLM,msrType); %Get delta X
    DeltaX = limitDelta(DeltaX, DxLimit); %Sanity check
    
    Fp = F;
    [F] = fn(T,I,X+DeltaX,msrType); %Evaluate at new pos
    N(1) = N(1)+1;
    plotCvg(X+DeltaX,X,F,Fp,verb,'FirstTry'); %Plot progress
    
    %Inner Loop---------------------------------------
    for c2=1:10,
      %Has function increased?
      if F > Fp,
        X = X+DeltaX;
        [F G H] = fn(T,I,X,msrType);
        N(2) = N(2)+1;
        plotCvg(X,[],F,[],verb,'SuccessfulTry'); %Plot progress
        break;
      end
      
      %Otherwise try another position
      l = 2*l;
      hLM = H .* (eye(length(X)) * l + 1); %Get LM Hessian
      if abs(rcond(hLM))<1e-3, hLM=eye(6); end; %Check if hLM is invertible
      
      DeltaX = getDelta(G,hLM,msrType); %Get delta X
      DeltaX = limitDelta(DeltaX, DxLimit); %Sanity check
      
      [F] = fn(T,I,X+DeltaX,msrType); %Evaluate at new pos
      N(1) = N(1)+1;
      plotCvg(X+DeltaX,X,F,Fp,verb,'OtherTries'); %Plot progress
    end %--------------------------------------------

    %Brent minimise
    %fnBrent(T,I,msrType);
    %[Xbr,Fbr] = brentWrap( @fnBrent, X+DeltaX, G );
    %X = X + Xbr*G;
    %F = -Fbr;
    
    %Store history 
    History.x = [History.x; X(:)']; 
    History.y = [History.y; F];

    if F-Fp<1e-4, break; end %Terminate based on delta F?
    if max(abs(DeltaX))<1e-4, break; end %Terminate based on delta X?
  end %==============================================
  
  plotCvg(X,[],F,[],verb,'end'); %Plot progress
  
%---------------------------------------------------------------------
function delta = getDelta(G,H,msrType)
% GETDELTA - Find the warp parameter update given the gradient and hessian
%   
  switch msrType{1}
   case 'mifwds'
    delta = G/H;
   case 'mibwds'
    delta = invertWarp( G/H );
   case 'ssdfwds'
    delta = G/H;
   case 'ssdbwds'
    delta = invertWarp( G/H );
    %delta = -G/H;
  end
  
%---------------------------------------------------------------------
function [f] = fnBrent(T,I,iType)
  
  persistent gFn; %Holds the parameters used by the function
  
  if nargin>1
    %disp('Setting up global fn parameters (gFn)');
    %disp('3 inputs used: tmp, img, param');
    
    if nargin < 3
      disp('Error: not enough arguments. 3 needed.');
      return;
    end
    
    gFn.tmp = T;
    gFn.img = I;
    gFn.type = iType;
    return;
  end

  X = T;
  
  %Check for the existence of gFn
  if ~exist('gFn','var')
      disp('Error: Create gFn first, by calling fn(T,I,P)');
      return;
  end

  fMsr = gFn.type{1};
  if numel(gFn.type) >= 3, PZ = gFn.type{3}; else PZ=1; end

  switch fMsr
   case 'mifwds'    
    f = mifwds(gFn.tmp,gFn.img,X,PZ); f=-f;
   case 'mibwds'
    f = mibwds(gFn.tmp,gFn.img,X,PZ); f=-f;
   case 'ssdfwds'
    f = ssdfwds(gFn.tmp,gFn.img,X); %Invert value 'cos we maximise
   case 'ssdbwds'
    f = ssdbwds(gFn.tmp,gFn.img,X); %Invert value 'cos we maximise
  end  
  
%---------------------------------------------------------------------
function [f,g,h] = fn(T,I,X,msrType)
% FN - Measure the function value at a particular position using a value
% selected by 't'
  
  fMsr = msrType{1};
  fMsrOrder = msrType{2};
  if numel(msrType) >= 3, PZ = msrType{3}; else PZ=1; end
    
  switch nargout
   case 1
    switch fMsr
     case 'mifwds'    
      f = mifwds(T,I,X,PZ);
     case 'mibwds'
      f = mibwds(T,I,X,PZ);
     case 'ssdfwds'
      f = ssdfwds(T,I,X); %Invert value 'cos we maximise
      f=-f;
     case 'ssdbwds'
      f = ssdbwds(T,I,X); %Invert value 'cos we maximise
      f=-f;
    end
   case 3
    switch fMsr
     case 'mifwds'
      switch fMsrOrder
       case 1
        [f g h] = mifwds(T,I,X,PZ);
       case 2
        [f g h h2] = mifwds(T,I,X,PZ);
        h = h+h2;
      end
      
     case 'mibwds'
      switch fMsrOrder
       case 1
        if ~exist('bwdsFirst','var')
          persistent mibwdsFirst miconstH;
          mibwdsFirst = 1;
        end
        if mibwdsFirst==1
          %[f g h h2 h3] = mibwds(T,I,X,PZ);
          %h=h+h2+h3; %Do we use full hess to start
          [f g h h2] = mibwds(T,I,X,PZ);
          h=h+h2; %Do we use full hess to start
          %[f g h] = mibwds(T,I,X,1);

          mibwdsFirst=0;
          miconstH = h;
        else
          [f g] = mibwds(T,I,X,PZ);
          h = miconstH;
        end
       case 1.5
        [f g h h2] = mibwds(T,I,X,PZ);
        h=h+h2;
       case 2
        [f g h h2 h3] = mibwds(T,I,X,PZ);
        h = h+h2+h3;
      end
      
     case 'ssdfwds'
      switch fMsrOrder
       case 1
        [f g h] = ssdfwds(T,I,X);
       case 2
        [f g h h2] = ssdfwds(T,I,X);
        h = h+h2;
      end
      f=-f; g=-g;
      
     case 'ssdbwds'
      switch fMsrOrder
       case 1
        if ~exist('bwdsFirst','var')
          persistent ssdbwdsFirst ssdconstH;
          ssdbwdsFirst = 1;
        end
        if ssdbwdsFirst==1
          %[f g h h2] = ssdbwds(T,I,X,PZ);
          %h=h+h2; %Do we use full hess to start
          [f g h] = ssdbwds(T,I,X);

          ssdbwdsFirst=0;
          ssdconstH = h;
        else
          [f g] = ssdbwds(T,I,X);
          h = ssdconstH;
        end
       case 2
        [f g h h2] = ssdbwds(T,I,X);
        h = h+h2;
      end
      f=-f; g=-g;
      
    end
  end
  
%---------------------------------------------------------------------
function plotCvg(X,Xp,F,Fp,verb,t)
% PLOTCVG - Plot the route the convergence algorithm takes
%   
  if length(verb)==0, return; end
  
  switch t
   case 'begin'
    subplot(1,3,1); hold off; mesh(verb.X1,verb.X2,verb.Y12);
    hold on; alpha(0.1); colormap summer;
    plot3(X(1),X(2),F,'ro');
    xlabel('v_1'); ylabel('v_2');
    subplot(1,3,2); hold off; mesh(verb.X3,verb.X4,verb.Y34);
    hold on; alpha(0.1); colormap summer;
    plot3(X(3),X(4),F,'ro');
    xlabel('v_3'); ylabel('v_4');
    subplot(1,3,3); hold off; mesh(verb.X5,verb.X6,verb.Y56);
    hold on; alpha(0.1); colormap summer;
    plot3(X(5),X(6),F,'ro');
    xlabel('v_5'); ylabel('v_6');
    
   case 'FirstTry'
    subplot(1,3,1); plot3(Xp(1),Xp(2),Fp,'k*');
    line([Xp(1) X(1)],[Xp(2) X(2)],[Fp F],'Color',[0 0 1]);
    subplot(1,3,2); plot3(Xp(3),Xp(4),Fp,'k*');
    line([Xp(3) X(3)],[Xp(4) X(4)],[Fp F],'Color',[0 0 1]);
    subplot(1,3,3); plot3(Xp(5),Xp(6),Fp,'k*');
    line([Xp(5) X(5)],[Xp(6) X(6)],[Fp F],'Color',[0 0 1]);
    
   case 'SuccessfulTry'
    subplot(1,3,1); plot3(X(1),X(2),F,'k+');
    subplot(1,3,2); plot3(X(3),X(4),F,'k+');
    subplot(1,3,3); plot3(X(5),X(6),F,'k+');
   
   case 'OtherTries'
    subplot(1,3,1); 
    line([Xp(1) X(1)],[Xp(2) X(2)],[Fp F],'Color',[0 0 1]);
    subplot(1,3,2); 
    line([Xp(3) X(3)],[Xp(4) X(4)],[Fp F],'Color',[0 0 1]);
    subplot(1,3,3); 
    line([Xp(5) X(5)],[Xp(6) X(6)],[Fp F],'Color',[0 0 1]);
    
   case 'end'
    subplot(1,3,1); 
    axis([verb.X1(1,1) verb.X1(end,1) verb.X2(1,1) verb.X2(1,end)]);
    plot3(X(1),X(2),F,'rs');
    subplot(1,3,2); 
    axis([verb.X3(1,1) verb.X3(end,1) verb.X4(1,1) verb.X4(1,end)]);
    plot3(X(3),X(4),F,'rs');
    subplot(1,3,3); 
    axis([verb.X5(1,1) verb.X5(end,1) verb.X6(1,1) verb.X6(1,end)]);
    plot3(X(5),X(6),F,'rs');
   
  end
  
%---------------------------------------------------------------------
function oDelta = limitDelta(iDelta,iLimit)
% LIMITDELTA - Limit the shift induced in delta in a single step
%   a  
  [propShift idx] = max( abs(iDelta(:) ./ iLimit(:)) );
  if( propShift > 1.0 )
    oDelta = iDelta / propShift;
  else
    oDelta = iDelta;
  end
  