// ********************************************************************
// mibwds
//
// Inverse compositional implementation for MI measurement
//
//  Author: N Dowson
//    Date: 24 Mar 06, REvised to correct off by 1 error for even sized templates
// Version: 0.6, 
// ********************************************************************

//Std c++
#include <iostream>

#include "mex.h"
#define cimg_plugin  "cimgmatlabNick.h"
#include <CImg-1-17/CImg.h>

using namespace cimg_library;

//********************************************************************
void mexErr(char* err, int nlhs, mxArray *plhs[] )
{
  std::cout << "\nMIBWDS Error: " << err << "\n" << " "
	    << "\n\nAborting...\n";
  //Create some output results
  for( int i=0; i<nlhs; i++ )
    plhs[i] = mxCreateNumericMatrix( 1, 1, mxSINGLE_CLASS, mxREAL);
}

//********************************************************************
CImg<double> jointHist( CImg<double>& if2a, CImg<double>& if2b, 
			int amax=0, int bmax=0 );
CImg<double> jointHist( CImg<double>& if2a, CImg<double>& if2b,
		       CImg<double>& if2val,
		       int amax=0, int bmax=0  );
CImg<double> singleHist( CImg<double>& if2a, int amax=0 );
CImg<double> singleHist( CImg<double>& if2a, CImg<double>& if2val, int amax=0 );
CImg<double> sumDim( CImg<double>& if2img, int iiDim=1 );
double entropy( CImg<double>& if2hist );
CImg<double> region( CImg<double>& if2tmp, CImg<double>& if2ref,  
		    CImg<double>& if2warp );
CImgl<double> imgXY( CImg<double>& if2tmp );
void outputImg( CImg<double> &if2a, char* iName );
CImgl<double> ImgListMul( CImgl<double> &ilf2a, int iima, int iina,
			 CImgl<double> &ilf2b, int iimb, int iinb );
CImgl<double> getdWdV( CImgl<double> &ilf2xy );
CImgl<double> ImgListTrans( CImgl<double> &ilf2a, int iima, int iina );


//********************************************************************
void mexFunction(int nlhs, mxArray *plhs[], int nrhs,
                 const mxArray *prhs[])
{
  //==================================================================
  //Initialisation
  using namespace std;
  
  //Check there are 3-4 inputs
  if( nrhs!=3 && nrhs!=4 ) 
    { mexErr( "3-4 inputs required.", nlhs, plhs ); return; }
 
  //Check there are 1-4 outputs
  if( (nlhs>5) )
    { mexErr( "1-5 outputs required", nlhs, plhs ); return; }
  
  //Read the inputs
  CImg<double> ff2tmp ( prhs[0], false, mxGetClassID(prhs[0]) == mxDOUBLE_CLASS ); //Template image
  CImg<double> ff2ref ( prhs[1], false, mxGetClassID(prhs[1]) == mxDOUBLE_CLASS  ); //Reference image
  CImg<double> ff2warp( prhs[2] ); //The Affine warp
  CImg<double> ff2parz = nrhs>=4 ? CImg<double>::CImg( prhs[3] ) 
    : CImg<double>::CImg(1,1,1,1,1.0); //The parzen window

  //==================================================================
  //Body
  int fiNt = (int) CImgStats( ff2tmp ).max + 1;
  int fiNr = (int) CImgStats( ff2ref ).max + 1;

//   cout<<"\nNt,r: "<<fiNt<<","<<fiNr;
  //--------------------------------------------------------------
  //Function
  
  //Extract region in ff2ref corresponding to warped ff2tmp
  CImg<double> ff2region = region(ff2tmp,ff2ref,ff2warp);
  //Generate a joint histogram and parzen window it
  //Increase histogram because we're convolving
  int fiNtDelta = ff2parz.dimx()-1+2; fiNt += fiNtDelta;
  int fiNrDelta = ff2parz.dimy()-1;   fiNr += fiNrDelta;
  ff2tmp += fiNtDelta/2;
  ff2region += fiNrDelta/2; //Shift samples 

  CImg<double> ff2Hist = jointHist( ff2tmp,ff2region,fiNt,
				   fiNr ).convolve( ff2parz,0 );

  //Obtain the mutual information
  CImg<double> ff2HistR = sumDim(ff2Hist,1);
  CImg<double> ff2HistT = sumDim(ff2Hist,2);
  double ofMI = entropy( ff2HistR ) + entropy( ff2HistT ) - entropy( ff2Hist );

  //Output result
  plhs[0] = CImg<double>( 1,1,1,1,ofMI).toMatlab();
  if( nlhs <= 1 ) return;

  //--------------------------------------------------------------
  //Gradient

  //Get 1st derivative parzen window
  CImg<double> ff2parzDt1( ff2parz.dimx()+1, ff2parz.dimy(), 1,1, 0.0 );
  //CImg<double> ff2parzMt1( ff2parz.dimx()+1, ff2parz.dimy(), 1,1, 0.0 );
  CImg<double> ff2Dt1(2,1,1,1,-1.0 ); ff2Dt1[1]= 1.0; //-1 1
  CImg<double> ff2Mt1(2,1,1,1, 0.5 );                 //.5 .5
  ff2parzDt1.draw_image( ff2parz,1,0 ).convolve( ff2Dt1,0 );
  //ff2parzMt1.draw_image( ff2parz,1,0 ).convolve( ff2Mt1,0 );

  //Extract region from gradient image
  CImgl<double> flf2tmpGrad = ff2tmp.get_gradientXY( 1 );

  //Generate warp derivative components
  CImgl<double> flf2tmpXY = imgXY( ff2tmp );
  //Compute Jacobian Image
  CImgl<double> flf2JacImg( flf2tmpXY[0].get_mul(flf2tmpGrad[0]),
			   flf2tmpXY[0].get_mul(flf2tmpGrad[1]),
			   flf2tmpXY[1].get_mul(flf2tmpGrad[0]),
			   flf2tmpXY[1].get_mul(flf2tmpGrad[1]),
			   flf2tmpGrad[0], flf2tmpGrad[1] );

  //Compute Jacobian Histogram
  CImgl<double> flf2JacHist( 6, fiNt, fiNr );
  for( int n=0; n<6; n++ )
    flf2JacHist[n] = jointHist( ff2tmp,ff2region,flf2JacImg[n], 
				fiNt,fiNr ).convolve(ff2parzDt1,0);
  //Compute gradient MI
  CImg<double> of2MIjac(1,6,1,1,0.0);
  CImg<double> ff2histMt1 = ff2Hist.get_convolve(ff2Mt1,0);
  CImg<double> ff2histMt1R= sumDim( ff2histMt1, 1 );
  CImg<double> ff2histMt1T= sumDim( ff2histMt1, 2 ); 
 
  for( int n=0; n<6; n++ ) {
    for( int fiy=0; fiy<ff2histMt1.dimy(); fiy++ ) {
      for( int fix=0; fix<ff2histMt1.dimx(); fix++ ) {
	if( ff2histMt1(fix,fiy) != 0 )
	  {
	    of2MIjac[n] += flf2JacHist[n](fix,fiy) * 
	      (1.0 + log(ff2histMt1(fix,fiy) / 
			 (ff2histMt1R(0,fiy) * ff2histMt1T(fix,0)) ));
	  }}}}  

  //Output result  
  plhs[1] = of2MIjac.toMatlab();
  if( nlhs <= 2 ) return;

  //--------------------------------------------------------------
  //Hessian: 1st order part

  CImg<double> ff2parzT = sumDim(ff2parz,2);
  //Get 2nd derivative parzen window
  CImg<double> ff2parzTDt2( ff2parzT.dimx()+2, 1,1,1, 0.0 );
  //CImg<double> ff2parzTMt2( ff2parzT.dimx()+2, 1,1,1, 0.0 );
  CImg<double> ff2parzTDth( ff2parzT.dimx()+2, 1,1,1, 0.0 );
  CImg<double> ff2Dt2(3,1,1,1,-1.0 ); ff2Dt2[1]= 2.0; //-1 2 -1
  CImg<double> ff2Mt2(3,1,1,1, 0.25); ff2Mt2[1]= 0.5; //.25 .5 .25
  CImg<double> ff2Dth(3,1,1,1, 0.0); ff2Dth[0]= -0.5; ff2Dth[2]= 0.5;//-.5 0 .5
  ff2parzTDt2.draw_image( ff2parzT,1,0 ).convolve( ff2Dt2,0 );
  //ff2parzTMt2.draw_image( ff2parzT,1,0 ).convolve( ff2Mt2,0 );
  ff2parzTDth.draw_image( ff2parzT,1,0 ).convolve( ff2Dth,0 );

  // (del T dwdv)^2
  CImgl<double> flf2vTdwdvSqd = ImgListMul( flf2JacImg,6,1, flf2JacImg,1,6 );

  // (Del^2 T) (dwdv)^2
  CImgl<double> flf2tmpDDX = flf2tmpGrad[0].get_gradientXY(1);
  CImgl<double> flf2tmpDDY = flf2tmpGrad[1].get_gradientXY(1);  
  CImgl<double> flf2tmpDD( flf2tmpDDX[0], flf2tmpDDX[1],
			  flf2tmpDDY[0], flf2tmpDDY[1] );
  CImgl<double> flf2dwdv = getdWdV( flf2tmpXY );
  CImgl<double> flf2dwdvT = ImgListTrans( flf2dwdv,2,6 );

  //Get 2nd order Histogram Hessian
//   CImg<double> ff2conv1(3,1,1,1,-1.0); ff2conv1[1] = 2.0;
//   CImg<double> ff2conv2(3,1,1,1, 0.0); ff2conv2[0] = -.5;  ff2conv2[2] = .5;
//   CImg<double> ff2conv3(3,1,1,1,0.25); ff2conv3[1] = 0.5;
  
  //Compute
  CImgl<double> flf2tmp = ImgListMul( flf2dwdvT,6,2, flf2tmpDD,2,2 );
  CImgl<double> flf2vvTdwdvSqd = ImgListMul( flf2tmp,6,2, flf2dwdv,2,6 );

  CImgl<double> flf2Hist1(36,fiNt,1,1,1);
  CImgl<double> flf2Hist2(36,fiNt,1,1,1);

  for( int n=0; n<36; n++ )
    {
      flf2Hist1[n] = singleHist( ff2tmp,flf2vTdwdvSqd[n],
				fiNt ).convolve( ff2parzTDt2,0 );
      flf2Hist2[n] = singleHist( ff2tmp,flf2vvTdwdvSqd[n],
				fiNt ).convolve( ff2parzTDth,0 );
    }

  CImg<double> ff2pt = singleHist( ff2tmp, fiNt ).convolve(ff2Mt2,0);
  CImgl<double> flf2ptD( 6,ff2pt.dimx(),1,1,1 );
  for( int n=0; n<6; n++ )
    {
      flf2ptD[n] = singleHist( ff2tmp, flf2JacImg[n], 
			       fiNt ).convolve( ff2parzTDth,0 );
    }
  CImgl<double> flf2ptDsqd = ImgListMul( flf2ptD,6,1, flf2ptD,1,6 );

  //MI hessian 1
  CImg<double> of2MIhess1(6,6,1,1,0.0);
  for( int n=0; n<36; n++ ) {
    for( int fix=0; fix<ff2pt.dimx(); fix++ ) {
      if( ff2pt(fix) !=0 ) 
	{
	  of2MIhess1[n] -= flf2ptDsqd[n](fix)/ff2pt(fix) 
	    + flf2Hist1[n](fix) + flf2Hist1[n](fix);
	}}}

  //Output result
  plhs[2] = of2MIhess1.toMatlab();
  if( nlhs <= 3 ) return;

  //--------------------------------------------------------------
  //Hessian: 1st order part R dependent part
  
  CImgl<double> flf2histJacSqd = ImgListMul( flf2JacHist,6,1, flf2JacHist,1,6 );

  //MI hessian 2
  CImg<double> of2MIhess2(6,6,1,1,0.0);

  for( int n=0; n<36; n++ ) {
    for( int fiy=0; fiy<ff2histMt1.dimy(); fiy++ ) {
      for( int fix=0; fix<ff2histMt1.dimx(); fix++ ) {
	if( ff2histMt1(fix,fiy)!=0 )
	  {
	    of2MIhess2[n] += flf2histJacSqd[n](fix,fiy) / ff2histMt1(fix,fiy);
	  }}}}

  //Output result
  plhs[3] = of2MIhess2.toMatlab();
  if( nlhs <= 4 ) return;

  //--------------------------------------------------------------
  //Hessian: 2nd order part
  
  //Get 2nd derivative parzen window
  CImg<double> ff2parzDt2( ff2parz.dimx()+2, 1,1,1, 0.0 );
  //CImg<double> ff2parzMt2( ff2parz.dimx()+2, 1,1,1, 0.0 );
  CImg<double> ff2parzDth( ff2parz.dimx()+2, 1,1,1, 0.0 );
  ff2parzDt2.draw_image( ff2parz,1,0 ).convolve( ff2Dt2,0 );
  //ff2parzMt2.draw_image( ff2parz,1,0 ).convolve( ff2Mt2,0 );
  ff2parzDth.draw_image( ff2parz,1,0 ).convolve( ff2Dth,0 );

  CImgl<double> flf2Hist5(36,fiNr,fiNt,1,1);
  CImgl<double> flf2Hist6(36,fiNr,fiNt,1,1);
  for( int n=0; n<36; n++ )
    {
      flf2Hist5[n] = jointHist( ff2tmp, ff2region, flf2vTdwdvSqd[n],
				fiNt,fiNr ).convolve( ff2parzDt2,0 ); //-1 2 -1
      flf2Hist6[n] = jointHist( ff2tmp, ff2region, flf2vvTdwdvSqd[n],
				fiNt,fiNr ).convolve( ff2parzDth,0 );//-.5 0 .5
    }

  //Blur the joint hist again
  CImg<double> ff2histMt2 = ff2Hist.get_convolve(ff2Mt2,0);
  CImg<double> ff2histRMt2 = sumDim(ff2histMt2,1);
  CImg<double> ff2histTMt2 = sumDim(ff2histMt2,2);

  //MI hessian 3
  CImg<double> of2MIhess3(6,6,1,1,0.0);

  for( int n=0; n<36; n++ ) {
    for( int fiy=0; fiy<ff2histMt2.dimy(); fiy++ ) {
      for( int fix=0; fix<ff2histMt2.dimx(); fix++ ) {
	if( ff2histMt2(fix,fiy)!=0 )
	  {
	    of2MIhess3[n] += (flf2Hist5[n](fix,fiy) + flf2Hist6[n](fix,fiy)) * 
	      ( 1.0+ log( ff2histMt2(fix,fiy) / ( ff2histRMt2(0,fiy)*
						  ff2histTMt2(fix,0) )));
	  }}}}

  //Output result
  plhs[4] = of2MIhess3.toMatlab();
}

 
//==================================================================
//Get the joint histogram of two images.
//Assumes the sizes of the 2 images are the same
//Assumes that amax and bmax are correctly sized
CImg<double> jointHist( CImg<double>& if2a, CImg<double>& if2b, 
			int amax, int bmax )
{
  double* fpf1 = if2a.ptr(), *fpf2 = if2b.ptr();
  double* fpfend = fpf1 + if2a.size();
  double ffProb = 1.0/if2a.size();

  //Do we need to assign sizes to the histogram
  if( amax==0 ) amax = (int) CImgStats( if2a ).max + 1;
  if( bmax==0 ) bmax = (int) CImgStats( if2b ).max + 1;

  CImg<double> of2Hist( amax,bmax,1,1, 0.0 );

  //Populate the histogram
  while( fpf1 < fpfend )
    {
      of2Hist( (int) *fpf1, (int) *fpf2 )+=ffProb;
      fpf1++; fpf2++;
    }

  return of2Hist;
}

//==================================================================
//Get the joint histogram of two images - weighting samples by if2val
//Assumes the sizes of the 2 images are the same
//Assumes that amax and bmax are correctly sized
CImg<double> jointHist( CImg<double>& if2a, CImg<double>& if2b,
			CImg<double>& if2val,
			int amax, int bmax )
{
  double *fpf1 = if2a.ptr(), *fpf2 = if2b.ptr(), *fpfval = if2val.ptr();
  double *fpfend = fpf1 + if2a.size();

  //Do we need to assign sizes to the histogram
  if( amax==0 ) amax = (int) CImgStats( if2a ).max + 1;
  if( bmax==0 ) bmax = (int) CImgStats( if2b ).max + 1;

  CImg<double> of2Hist( amax,bmax,1,1, 0.0 );

  //Populate the histogram
  while( fpf1 < fpfend )
    {
      of2Hist( (int) *fpf1, (int) *fpf2 ) += *fpfval;
      fpf1++; fpf2++; fpfval++;
    }
  of2Hist /= if2a.size();

  return of2Hist;
}

//==================================================================
//Get the joint histogram of two images.
//Assumes the sizes of the 2 images are the same
//Assumes that amax and bmax are correctly sized
CImg<double> singleHist( CImg<double>& if2a, int amax )
{
  double* fpf1 = if2a.ptr();
  double* fpfend = fpf1 + if2a.size();
  double ffProb = 1.0/if2a.size();

  //Do we need to assign sizes to the histogram
  if( amax==0 ) amax = (int) CImgStats( if2a ).max + 1;

  CImg<double> of2Hist( amax,1,1,1, 0.0 );

  //Populate the histogram
  while( fpf1 < fpfend )
    {
      of2Hist[ (int) *fpf1 ] += ffProb;
      fpf1++;
    }

  return of2Hist;
}

//==================================================================
//Get the joint histogram of two images - weighting samples by if2val
//Assumes the sizes of the 2 images are the same
//Assumes that amax and bmax are correctly sized
CImg<double> singleHist( CImg<double>& if2a, CImg<double>& if2val, int amax )
{
  double *fpf1 = if2a.ptr(), *fpfval = if2val.ptr();
  double *fpfend = fpf1 + if2a.size();

  //Do we need to assign sizes to the histogram
  if( amax==0 ) amax = (int) CImgStats( if2a ).max + 1;

  CImg<double> of2Hist( amax,1,1,1, 0.0 );

  //Populate the histogram
  while( fpf1 < fpfend )
    {
      of2Hist[ (int) *fpf1 ] += *fpfval;
      fpf1++; fpfval++;
    }
  of2Hist /= if2a.size();

  return of2Hist;
}

//==================================================================
//Sums a given image along a particular dimension
CImg<double> sumDim( CImg<double>& if2img, int iiDim )
{
  //Which dimension are we summing along
  int fid[4], fids[4];
  fid[0] = if2img.dimx(); fid[1] = if2img.dimy();
  fid[2] = if2img.dimz(); fid[3] = if2img.dimv();
  for( int i=0; i<4; i++ ) fids[i] = fid[i];
  if( iiDim < 1 ) iiDim=1;
  if( iiDim > 4 ) iiDim=1;
  iiDim--;
  fids[iiDim] = 1;

  double* fpfimg = if2img.ptr(); //Pointer to img

  CImg<double> of2ImgSum( fids[0], fids[1], fids[2], fids[3], 0.0 );

  for( int fiv=0; fiv<if2img.dimv(); fiv++ ) 
    { 
      int fiv2 = iiDim==3 ? 0 : fiv;
      for( int fiz=0; fiz<if2img.dimz(); fiz++ ) 
	{
	  int fiz2 = iiDim==2 ? 0 : fiz;
	  for( int fiy=0; fiy<if2img.dimy(); fiy++ ) 
	    {
	      int fiy2 = iiDim==1 ? 0 : fiy;
	      for( int fix=0; fix<if2img.dimx(); fix++ )
		{
		  int fix2 = iiDim==0 ? 0 : fix;
		  of2ImgSum( fix2, fiy2, fiz2, fiv2 ) += *fpfimg;
		  fpfimg++;
		}}}}

  return of2ImgSum;
}

//==================================================================
//Measure the entropy of a given histogram.
//Assumes everything sums to one.
double entropy( CImg<double>& if2hist )
{
  double *fpfPixel = if2hist.ptr(), *fpfEnd = fpfPixel+if2hist.size();
  double ofEntropy = 0.0;

  while( fpfPixel < fpfEnd )
    {
      double v = *fpfPixel;

      if( v>0.0 ) 
	ofEntropy += v*log(v);

      fpfPixel++;
    }
  return -ofEntropy;
}

//==================================================================
//Sums a given image along a particular dimension
//Assumes images are 2D - does linear interpolation
CImg<double> region( CImg<double>& if2tmp, CImg<double>& if2ref,  
		    CImg<double>& if2warp )
{
  CImg<double> of2region( if2tmp, false );
  of2region=0.0;
  double *fpfPixel = of2region.ptr();

  //Off by one error for even sized templates
  //int fiyA = -(if2tmp.dimy()-1)/2, fiyE = (if2tmp.dimy()-1)/2;
  //int fixA = -(if2tmp.dimx()-1)/2, fixE = (if2tmp.dimx()-1)/2;
  int fiyA = (int) (-(if2tmp.dimy()-1)/2);
  int fiyE = fiyA+(int)if2tmp.dimy();
  int fixA = (int) (-(if2tmp.dimx()-1)/2);
  int fixE = fixA+(int)if2tmp.dimx();

  for( int fiy=fiyA; fiy<fiyE; fiy++ )
    for( int fix=fixA; fix<fixE; fix++ )
      {
	double ffxw = (1+if2warp[0])*fix +    if2warp[2] *fiy + if2warp[4];
	double ffyw =    if2warp[1] *fix + (1+if2warp[3])*fiy + if2warp[5];
	(*fpfPixel) = if2ref.linear_pix2d( static_cast<float>(ffxw),
    static_cast<float>(ffyw) );

//  	std::cout<<"\ntx,y rx,y v: "<< fix<<","<<fiy<<" "
//  		 <<ffxw<<","<<ffyw<<" "<<(*fpfPixel);
	fpfPixel++;
      }

  return of2region;
}

//==================================================================
//Sums a given image along a particular dimension
//Assumes images are 2D - does linear interpolation
CImgl<double> imgXY( CImg<double>& if2tmp )
{
  CImgl<double> olf2XY( 2, if2tmp.dimx(), if2tmp.dimy() ); 
//   olf2XY[0]=0.0;
//   olf2XY[1]=0.0;

  double half = olf2XY[0].dimx()/2;
  cimg_mapXY(olf2XY[0],x,y) olf2XY[0](x,y)=x-half;

  half = olf2XY[0].dimy()/2;
  cimg_mapXY(olf2XY[1],x,y) olf2XY[1](x,y)=y-half;
  return olf2XY;
}

//==================================================================
//Output an image in matrix form
void outputImg( CImg<double> &if2a, char* iName )
{
  using namespace std;
  cout<<"\n"<<iName<<" = reshape([";
  for( unsigned int i=0; i<if2a.size(); i++ ) cout<<if2a[i]<<" ";
  cout<<"],["<<if2a.dimx()<<" "<<if2a.dimy()<<"]);\n";
}

//==================================================================
//Multiply two matrices of images together
//Assumes images are all same size and that iina==iimb
CImgl<double> ImgListMul( CImgl<double> &ilf2a, int iima, int iina,
			 CImgl<double> &ilf2b, int iimb, int iinb )
{
  CImgl<double> olf2Result( iima*iinb, ilf2a[0].dimx(), ilf2a[0].dimy(),
			   ilf2a[0].dimz(), ilf2a[0].dimv(), 0.0 );
  
  for( int fim=0; fim<iima; fim++ ) {
    for( int fin=0; fin<iinb; fin++ ) {
      for( int fie=0; fie<iina; fie++ ) 
	{
	  //olf2Result(fim,fin) += ilf2a(fim,fie) * ilf2b(fie,fin);
	  olf2Result[fim+fin*iima] += //iima*iinb
	    ilf2a[fim+fie*iima].get_mul( //iima*iina
	    ilf2b[fie+fin*iimb] );  //iimb*iinb
	}}}      

  return olf2Result;
}

//==================================================================
//Multiply two matrices of images together
//Assumes images are all same size and that iina==iimb
CImgl<double> getdWdV( CImgl<double> &ilf2xy )
{
  CImgl<double> olf2dWdV( 12, ilf2xy[0].dimx(), ilf2xy[0].dimy(), 
			 ilf2xy[0].dimz(), ilf2xy[0].dimv() );

  olf2dWdV[ 0] = ilf2xy[0]; olf2dWdV[ 1] = 0.0 ;
  olf2dWdV[ 2] = 0.0;       olf2dWdV[ 3] = ilf2xy[0];
  olf2dWdV[ 4] = ilf2xy[1]; olf2dWdV[ 5] = 0.0 ;
  olf2dWdV[ 6] = 0.0;       olf2dWdV[ 7] = ilf2xy[1];
  olf2dWdV[ 8] = 1.0;       olf2dWdV[ 9] = 0.0 ;
  olf2dWdV[10] = 0.0;       olf2dWdV[11] = 1.0 ;
  
  return olf2dWdV;
}

//==================================================================
//Transpose a matrix of images
//Assumes images are all same size
CImgl<double> ImgListTrans( CImgl<double> &ilf2a, int iima, int iina )
{
  CImgl<double> olf2trans( ilf2a );
  
  for( int fim=0; fim<iina; fim++ ){
    for( int fin=0; fin<iima; fin++ )
      {
	//olf2trans(fim,fin) = ilf2a(fin,fim);
	olf2trans[fim+fin*iina] = //iina*iima
	  ilf2a[fin+fim*iima]; //iima*iina
      }}

  return olf2trans;
}
