// ********************************************************************
// ssdfwds
//
// Forwards additive implementation for MI measurement
//
//  Author: N Dowson
//    Date: 24 Mar 06, REvised to correct off by 1 error for even sized templates
// Version: 0.6, 
// ********************************************************************

//Std c++
#include <iostream>

#include "mex.h"
#define cimg_plugin  "cimgmatlabNick.h"
#include <CImg-1-17/CImg.h>

using namespace cimg_library;

//********************************************************************
void mexErr(char* err, int nlhs, mxArray *plhs[] )
{
  std::cout << "\nSSDFWDS Error: " << err << "\n" << " "
	    << "\n\nAborting...\n";
  //Create some output results
  for( int i=0; i<nlhs; i++ )
    plhs[i] = mxCreateNumericMatrix( 1, 1, mxSINGLE_CLASS, mxREAL);
}

//********************************************************************
CImg<double> jointHist( CImg<double>& if2a, CImg<double>& if2b, 
			int amax=0, int bmax=0 );
CImg<double> jointHist( CImg<double>& if2a, CImg<double>& if2b,
		       CImg<double>& if2val,
		       int amax=0, int bmax=0  );
CImg<double> sumDim( CImg<double>& if2img, int iiDim=1 );
double entropy( CImg<double>& if2hist );
CImg<double> region( CImg<double>& if2tmp, CImg<double>& if2ref,  
		    CImg<double>& if2warp );
CImgl<double> imgXY( CImg<double>& if2tmp );
void outputImg( CImg<double> &if2a, char* iName );
CImgl<double> ImgListMul( CImgl<double> &ilf2a, int iima, int iina,
			 CImgl<double> &ilf2b, int iimb, int iinb );
CImgl<double> getdWdV( CImgl<double> &ilf2xy );
CImgl<double> ImgListTrans( CImgl<double> &ilf2a, int iima, int iina );


//********************************************************************
void mexFunction(int nlhs, mxArray *plhs[], int nrhs,
                 const mxArray *prhs[])
{
  //==================================================================
  //Initialisation
  using namespace std;
  
  //Check there are 3-4 inputs
  if( nrhs!=3 ) 
    { mexErr( "3 inputs required.", nlhs, plhs ); return; }
 
  //Check there are 1-4 outputs
  if( (nlhs>4) )
    { mexErr( "1-4 outputs required", nlhs, plhs ); return; }
  
  //Read the inputs
  CImg<double> ff2tmp ( prhs[0], false, mxGetClassID(prhs[0]) == mxDOUBLE_CLASS ); //Template image
  CImg<double> ff2ref ( prhs[1], false, mxGetClassID(prhs[1]) == mxDOUBLE_CLASS  ); //Reference image
  CImg<double> ff2warp( prhs[2] ); //The Affine warp

  //==================================================================
  //Body

  //--------------------------------------------------------------
  //Function

  //Extract region in ff2ref corresponding to warped ff2tmp
  CImg<double> ff2region = region(ff2tmp,ff2ref,ff2warp);

  CImg<double> ff2Diff = ff2region - ff2tmp;
  double ofSSD = ff2Diff.get_mul(ff2Diff).sum();

  //Output result
  plhs[0] = CImg<double>( 1,1,1,1,ofSSD).toMatlab();
  if( nlhs <= 1 ) return;

  //--------------------------------------------------------------
  //Gradient

  //Extract region from gradient image
  CImgl<double> flf2refGrad = ff2ref.get_gradientXY( 1 );
  CImgl<double> flf2regionGrad( region(ff2tmp,flf2refGrad[0],ff2warp),
			       region(ff2tmp,flf2refGrad[1],ff2warp) );
  //Generate warp derivative components
  CImgl<double> flf2tmpXY = imgXY( ff2tmp );
  //Compute Jacobian Image
  CImgl<double> flf2JacImg( flf2tmpXY[0].get_mul(flf2regionGrad[0]),
			   flf2tmpXY[0].get_mul(flf2regionGrad[1]),
			   flf2tmpXY[1].get_mul(flf2regionGrad[0]),
			   flf2tmpXY[1].get_mul(flf2regionGrad[1]),
			   flf2regionGrad[0], flf2regionGrad[1] );

  //Compute gradient MI
  CImg<double> of2SSDjac(1,6,1,1,0.0);
 
  for( int n=0; n<6; n++ ) {
    of2SSDjac[n] = (flf2JacImg[n].get_mul(ff2Diff) ).sum();
  }
  
  //Output result  
  plhs[1] = of2SSDjac.toMatlab();
  if( nlhs <= 2 ) return;

  //--------------------------------------------------------------
  //Hessian: 1st order part

  //Compute 1st order Hessian of SSD
  CImg<double> of2SSDhess1(6,6,1,1,0.0);

  for( int m=0; m<6; m++ ) {
    for( int n=0; n<6; n++ ) {
      of2SSDhess1(m,n) = flf2JacImg[n].get_mul(flf2JacImg[m]).sum();
    }}
  //Output result
  plhs[2] = of2SSDhess1.toMatlab();
  if( nlhs <= 3 ) return;

  //--------------------------------------------------------------
  //Hessian: 2nd order part

  // (del R dwdv)^2
  CImgl<double> flf2vRdwdvSqd = ImgListMul( flf2JacImg,6,1, flf2JacImg,1,6 );
  // (Del^2 R) (dwdv)^2
  CImgl<double> flf2refDDX = flf2refGrad[0].get_gradientXY(1);
  CImgl<double> flf2refDDY = flf2refGrad[1].get_gradientXY(1);
  CImgl<double> flf2regionDD( region( ff2tmp, flf2refDDX[0], ff2warp ),
			     region( ff2tmp, flf2refDDX[1], ff2warp ),
			     region( ff2tmp, flf2refDDY[0], ff2warp ),
			     region( ff2tmp, flf2refDDY[1], ff2warp ) );

  CImgl<double> flf2dwdv = getdWdV( flf2tmpXY );  
  CImgl<double> flf2dwdvT = ImgListTrans( flf2dwdv,2,6 );

  //Compute
  CImgl<double> flf2tmp = ImgListMul( flf2dwdvT,6,2, flf2regionDD,2,2 );
  CImgl<double> flf2vvRdwdvSqd = ImgListMul( flf2tmp,6,2, flf2dwdv,2,6 );

  CImg<double> of2SSDhess2(6,6,1,1,0.0);
  for( int n=0; n<36; n++ ) {
    of2SSDhess2[n] = (flf2vvRdwdvSqd[n]+flf2vRdwdvSqd[n]).mul(ff2Diff).sum();
  }

  //Output result
  plhs[3] = of2SSDhess2.toMatlab();
}

 
//==================================================================
//Get the joint histogram of two images.
//Assumes the sizes of the 2 images are the same
//Assumes that amax and bmax are correctly sized
CImg<double> jointHist( CImg<double>& if2a, CImg<double>& if2b, 
			int amax, int bmax )
{
  double* fpf1 = if2a.ptr(), *fpf2 = if2b.ptr();
  double* fpfend = fpf1 + if2a.size();
  double ffProb = 1.0/if2a.size();

  //Do we need to assign sizes to the histogram
  if( amax==0 ) amax = (int) CImgStats( if2a ).max + 1;
  if( bmax==0 ) bmax = (int) CImgStats( if2b ).max + 1;

  CImg<double> of2Hist( amax,bmax,1,1, 0.0 );

  //Populate the histogram
  while( fpf1 < fpfend )
    {
      of2Hist( (int) *fpf1, (int) *fpf2 )+=ffProb;
      fpf1++; fpf2++;
    }

  return of2Hist;
}

//==================================================================
//Get the joint histogram of two images - weighting samples by if2val
//Assumes the sizes of the 2 images are the same
//Assumes that amax and bmax are correctly sized
CImg<double> jointHist( CImg<double>& if2a, CImg<double>& if2b,
			CImg<double>& if2val,
			int amax, int bmax )
{
  double *fpf1 = if2a.ptr(), *fpf2 = if2b.ptr(), *fpfval = if2val.ptr();
  double *fpfend = fpf1 + if2a.size();

  //Do we need to assign sizes to the histogram
  if( amax==0 ) amax = (int) CImgStats( if2a ).max + 1;
  if( bmax==0 ) bmax = (int) CImgStats( if2b ).max + 1;

  CImg<double> of2Hist( amax,bmax,1,1, 0.0 );

  //Populate the histogram
  while( fpf1 < fpfend )
    {
      of2Hist( (int) *fpf1, (int) *fpf2 ) += *fpfval;
      fpf1++; fpf2++; fpfval++;
    }
  of2Hist /= if2a.size();

  return of2Hist;
}

//==================================================================
//Sums a given image along a particular dimension
CImg<double> sumDim( CImg<double>& if2img, int iiDim )
{
  //Which dimension are we summing along
  int fid[4], fids[4];
  fid[0] = if2img.dimx(); fid[1] = if2img.dimy();
  fid[2] = if2img.dimz(); fid[3] = if2img.dimv();
  for( int i=0; i<4; i++ ) fids[i] = fid[i];
  if( iiDim < 1 ) iiDim=1;
  if( iiDim > 4 ) iiDim=1;
  iiDim--;
  fids[iiDim] = 1;

  double* fpfimg = if2img.ptr(); //Pointer to img

  CImg<double> of2ImgSum( fids[0], fids[1], fids[2], fids[3], 0.0 );

  for( int fiv=0; fiv<if2img.dimv(); fiv++ ) 
    { 
      int fiv2 = iiDim==3 ? 0 : fiv;
      for( int fiz=0; fiz<if2img.dimz(); fiz++ ) 
	{
	  int fiz2 = iiDim==2 ? 0 : fiz;
	  for( int fiy=0; fiy<if2img.dimy(); fiy++ ) 
	    {
	      int fiy2 = iiDim==1 ? 0 : fiy;
	      for( int fix=0; fix<if2img.dimx(); fix++ )
		{
		  int fix2 = iiDim==0 ? 0 : fix;
		  of2ImgSum( fix2, fiy2, fiz2, fiv2 ) += *fpfimg;
		  fpfimg++;
		}}}}

  return of2ImgSum;
}

//==================================================================
//Measure the entropy of a given histogram.
//Assumes everything sums to one.
double entropy( CImg<double>& if2hist )
{
  double *fpfPixel = if2hist.ptr(), *fpfEnd = fpfPixel+if2hist.size();
  double ofEntropy = 0.0;

  while( fpfPixel < fpfEnd )
    {
      double v = *fpfPixel;

      if( v>0.0 ) 
	ofEntropy += v*log(v);

      fpfPixel++;
    }
  return -ofEntropy;
}

//==================================================================
//Sums a given image along a particular dimension
//Assumes images are 2D - does linear interpolation
CImg<double> region( CImg<double>& if2tmp, CImg<double>& if2ref,  
		    CImg<double>& if2warp )
{
  CImg<double> of2region( if2tmp, false );
  of2region=0.0;
  double *fpfPixel = of2region.ptr();

  //Off by one error for even sized templates
  //int fiyA = -(if2tmp.dimy()-1)/2, fiyE = (if2tmp.dimy()-1)/2;
  //int fixA = -(if2tmp.dimx()-1)/2, fixE = (if2tmp.dimx()-1)/2;
  int fiyA = (int) (-(if2tmp.dimy()-1)/2);
  int fiyE = fiyA+(int)if2tmp.dimy();
  int fixA = (int) (-(if2tmp.dimx()-1)/2);
  int fixE = fixA+(int)if2tmp.dimx();

  for( int fiy=fiyA; fiy<fiyE; fiy++ )
    for( int fix=fixA; fix<fixE; fix++ )
      {
	double ffxw = (1+if2warp[0])*fix +    if2warp[2] *fiy + if2warp[4];
	double ffyw =    if2warp[1] *fix + (1+if2warp[3])*fiy + if2warp[5];
	(*fpfPixel) = if2ref.linear_pix2d( static_cast<float>(ffxw),
    static_cast<float>(ffyw) );

//  	std::cout<<"\ntx,y rx,y v: "<< fix<<","<<fiy<<" "
//  		 <<ffxw<<","<<ffyw<<" "<<(*fpfPixel);
	fpfPixel++;
      }

  return of2region;
}

//==================================================================
//Sums a given image along a particular dimension
//Assumes images are 2D - does linear interpolation
CImgl<double> imgXY( CImg<double>& if2tmp )
{
  CImgl<double> olf2XY( 2, if2tmp.dimx(), if2tmp.dimy() ); 
//   olf2XY[0]=0.0;
//   olf2XY[1]=0.0;

  double half = olf2XY[0].dimx()/2;
  cimg_mapXY(olf2XY[0],x,y) olf2XY[0](x,y)=x-half;

  half = olf2XY[0].dimy()/2;
  cimg_mapXY(olf2XY[1],x,y) olf2XY[1](x,y)=y-half;
  return olf2XY;
}

//==================================================================
//Output an image in matrix form
void outputImg( CImg<double> &if2a, char* iName )
{
  using namespace std;
  cout<<"\n"<<iName<<" = reshape([";
  for( unsigned int i=0; i<if2a.size(); i++ ) cout<<if2a[i]<<" ";
  cout<<"],["<<if2a.dimx()<<" "<<if2a.dimy()<<"]);\n";
}

//==================================================================
//Multiply two matrices of images together
//Assumes images are all same size and that iina==iimb
CImgl<double> ImgListMul( CImgl<double> &ilf2a, int iima, int iina,
			 CImgl<double> &ilf2b, int iimb, int iinb )
{
  CImgl<double> olf2Result( iima*iinb, ilf2a[0].dimx(), ilf2a[0].dimy(),
			   ilf2a[0].dimz(), ilf2a[0].dimv(), 0.0 );
  
  for( int fim=0; fim<iima; fim++ ) {
    for( int fin=0; fin<iinb; fin++ ) {
      for( int fie=0; fie<iina; fie++ ) 
	{
	  //olf2Result(fim,fin) += ilf2a(fim,fie) * ilf2b(fie,fin);
	  olf2Result[fim+fin*iima] += //iima*iinb
	    ilf2a[fim+fie*iima].get_mul( //iima*iina
	    ilf2b[fie+fin*iimb] );  //iimb*iinb
	}}}      

  return olf2Result;
}

//==================================================================
//Multiply two matrices of images together
//Assumes images are all same size and that iina==iimb
CImgl<double> getdWdV( CImgl<double> &ilf2xy )
{
  CImgl<double> olf2dWdV( 12, ilf2xy[0].dimx(), ilf2xy[0].dimy(), 
			 ilf2xy[0].dimz(), ilf2xy[0].dimv() );

  olf2dWdV[ 0] = ilf2xy[0]; olf2dWdV[ 1] = 0.0 ;
  olf2dWdV[ 2] = 0.0;       olf2dWdV[ 3] = ilf2xy[0];
  olf2dWdV[ 4] = ilf2xy[1]; olf2dWdV[ 5] = 0.0 ;
  olf2dWdV[ 6] = 0.0;       olf2dWdV[ 7] = ilf2xy[1];
  olf2dWdV[ 8] = 1.0;       olf2dWdV[ 9] = 0.0 ;
  olf2dWdV[10] = 0.0;       olf2dWdV[11] = 1.0 ;
  
  return olf2dWdV;
}

//==================================================================
//Transpose a matrix of images
//Assumes images are all same size
CImgl<double> ImgListTrans( CImgl<double> &ilf2a, int iima, int iina )
{
  CImgl<double> olf2trans( ilf2a );
  
  for( int fim=0; fim<iina; fim++ ){
    for( int fin=0; fin<iima; fin++ )
      {
	//olf2trans(fim,fin) = ilf2a(fin,fim);
	olf2trans[fim+fin*iina] = //iina*iima
	  ilf2a[fin+fim*iima]; //iima*iina
      }}

  return olf2trans;
}
