%[H,M] = trk(S,X,US,<M>)
%
% Synopsis: Tracks the feature at location X of frame 1 of sequence S
% with user selected optimisation algorithm OP and update strategy US.
% Outputs:
%    H - Tracking History
%   *M - Model storing state of current track. (* Also an output)
% Inputs:
%    S  - Sequence of Images (3D cube of pixels: x*y*time)
%    X  - Initial position of feature (2,3,4,6 values for translation,
%         euclidean, similarity or affine coordinates) + 2 for rect size
%    US - Verbosity (0/1 for quiet/verbose) and...
%         Presearch params (window size and density) and...
%         Update strategy: 1-6 for:
%         none, naive, strategic*, naive+real, strategic+real* and
%         forward-back respectively. (The * indicates that a second element
%         specifying a tolerence parameter is also required.)
%         Method to use for tracking - see lmmin help for details
%
%         e.g.

% Author:  Nicholas Dowson
% Revised: 12 Mar 06
%---------------------------------------------------------------------

%Warning this has LESS feautes than the p_trk version.

function [oHstry, ioM] = trk(iSeq, iX, iParam, ioM)
  figure(2); close(2);
  clear functions;
  if nargin < 3
    disp('trk error: Less than 3 parameters given. Exiting.');
    disp('Usage: [H] = trk(S,X,OP). Type "help trk" for more info.');
    return;
  end
  iX = iX(:);
  if numel(iX) ~= 8
    disp('trk error: iX must have 8 elements.');
    return;
  end
  
  %Parameters
  fGridRange = iParam{1}; %Pre search window size
  if length(fGridRange)==0, fGridRange = [3 1]; end
  fVerbosity = iParam{2}; %Verbosity
  fUpdStrat = iParam{3}; %Update strategy
  if numel(iParam)>=4, fUpdParam = iParam{4}; else fUpdParam=1; end
  if numel(iParam)>=5, fImgMul = iParam{5}; else fImgMul=1; end
  if numel(iParam)>=6, fTrk = iParam{6}; else fTrk={'mifwds' 1}; end
  
  % Initialisation
  fTmpSz = iX(7:8)*2+1;
  fX = iX(1:6);  
  oHstry  = [fX(:); 0; 1];

  %Put model together
  if nargin < 4
    ioM.T0 = affRegion( fTmpSz, iSeq(:,:,1), fX );
    ioM.TN = ioM.T0;
  end

  %Load the kernel from the first frame
  fI  = iSeq(:,:,1);
  fXo = zeros(size(fX)); %Get pixel offset

  fLMhst = [];
  trkStatus( fI, ioM, oHstry, fLMhst, fVerbosity, iParam );
  
  %For all the frames from frame 2
  for lFrm = [2:size(iSeq,3)]
    %Find position of best mutual information
    fIp = fI;
    fI = iSeq(:,:,lFrm);
    tb = fImgMul;
    
    if strcmp(fUpdStrat,'none') %No update--------------------------------
      fX = preSearch( ioM.T0/tb, fI/tb, fX, fGridRange ); %Presearch
      [fX,fY,fLMhst] = lmmin( ioM.T0/tb, fI/tb, fX,fTrk ); %Registration
      
    elseif strcmp(fUpdStrat,'naive')%Naive update-------------------------
      fX = preSearch( ioM.TN/tb, fI/tb, fX, fGridRange ); %Presearch
      [fX,fY,fLMhst] = lmmin( ioM.TN/tb, fI/tb, fX,fTrk ); %Registration
      ioM.TN = affRegion(fTmpSz, fI, fX );
      
    elseif strcmp(fUpdStrat,'strategic') %Strategic update----------------
      fX = preSearch( ioM.TN/tb, fI/tb, fX, fGridRange ); %Presearch
      [fX,fY,fLMhst1] = lmmin( ioM.TN/tb, fI/tb, fX,fTrk );
      fXpreCorr=fX;
      [fX,fY,fLMhst] = lmmin( ioM.T0/tb, fI/tb, fX,fTrk );
      fLMhst.x = [fLMhst1.x; fLMhst.x];
      fLMhst.y = [fLMhst1.y; fLMhst.y];
      %Block update if correction is large
      if sqrt(sum((fXpreCorr(5:6)-fX(5:6)).^2)) < fUpdParam
        ioM.TN = affRegion(fTmpSz, fI, fX );
      end
      
    elseif strcmp(fUpdStrat,'naiveZI') %Naive + ZI update----------------
      if ~isfield(ioM,'XtX0') ioM.XtX0 = zeros(size(fX)); end
      fX = preSearch( ioM.TN/tb, fI/tb, fX, iOP, preGrid, ix, iy ); %Presearch
      [fY,fXtI      ] = mexMIopt( ioM.TN, fI, fX, iOP );%Find best matching pos
      fX0I = fXtI - ioM.XtX0; %From Xt:I and Xt:X0, get X0:I
      fXniI= zeros(size(fX0I)); %Get zero interp position
      fXniI(ix:iy) = round(fX0I(ix:iy)); 
      ioM.TN = affRegion( kShp, fI, fXniI(ix:iy) ); %Update template
      ioM.XtX0 = fXniI - fX0I; %Get new Xt:X0
      fX = fXniI; %Back into main function variables names
      clear fX0I fXtI fXniI; %Clear extraneous variables
      
    elseif strcmp(fUpdStrat,'strategicZI') %Strategic + ZI update--------
      if ~isfield(ioM,'XtX0') ioM.XtX0 = zeros(size(fX)); end
      fX = preSearch( ioM.TN, fI, fX, iOP, preGrid, ix, iy ); %Presearch
      [fY,fXtI      ] = mexMIopt( ioM.TN, fI, fX,   iOP );
      eX0I = fXtI - ioM.XtX0; %From Xt:I and Xt:X0, estimate X0:I
      [fY,fX0I      ] = mexMIopt( ioM.T0, fI, eX0I, iOP ); %Correct fX0I
      %Block update if correction is large
      if sqrt(sum((eX0I(ix:iy)-fX0I(ix:iy)).^2)) < iUpd(3)
        fXniI= zeros(size(fX0I)); %Get zero interp position
        fXniI(ix:iy) = round(fX0I(ix:iy)); 
        ioM.TN = affRegion( kShp, fI, fXniI(ix:iy) ); %Update template
        ioM.XtX0 = fXniI - fX0I; %Get new Xt:X0
        fX = fXniI; %Back into main function variables names
        clear fXniI; %Clear extraneous variables
      else
        fX = fX0I;
      end
      clear eX0I fX0I fXtI;
      
    elseif strcmp(fUpdStrat,'naiveFwdBwd') %Naive Forwards-Back track------
      fXp = fX;
      fTp = ioM.TN; %Store previous template
      [fY,fX,fH,ioM.TN ] = mexMIopt( ioM.TN, fI , fX, iOP ); %Forwards track
      [fYb,fXb,fHb,fTb ] = mexMIopt( ioM.TN, fIp, fXp, iOP ); %Backwards track
      fX=fX-1.0*(fXp-fXb);
      ioM.TN = affRegion(kShp,fI,fX);       
      clear fYb fXb fHb fTb;
    else
      disp(['trk error: update strategy must be ''none'' ''naive'' '...
            ' ''strategic'' ''naiveZI'' ''strategicZI'' naiveFwdBwd'' ']);
      return;
    end
      
    oHstry = [oHstry [fX(:); fY; lFrm] ];
    trkStatus( fI, ioM, oHstry, fLMhst, fVerbosity, iParam );

  end
   
%---------------------------------------------------------------------
function oX = preSearch(T,I,iX,fGridRange)
  
  iX = iX(:);
  t = [-fGridRange(1):fGridRange(2):fGridRange(1)];
  if length(t) ~= 0
    %Pre-search 1
    [x y] = ndgrid(t,t);
    fGrid = [repmat(iX(1:4),1,numel(x)); iX(5)+x(:)'; iX(6)+y(:)'];
    
    fY = zeros(numel(x),1);    
    for c1=1:numel(x)
      fY(c1) = mifwds( T, I, fGrid(:,c1),1 );
    end

    [t idx] = min(fY);
    oX = fGrid(:,idx(1));
  else
    oX = iX;
  end

%---------------------------------------------------------------------
function trkStatus(fI,iMdl,iHist,iOptHist,iVerbosity,iParam)
% TRKSTATUS - Display the current tracking status
 
  persistent totalTime;
  if length(totalTime)==0
    totalTime=0; 
    tic; 
  end;
  tTime=toc;
  totalTime = totalTime+tTime;

  if numel(iOptHist)==0, iOptHist.x=[]; iOptHist.y=[]; end;
  if ~strcmp(iVerbosity,'verbose'), 
    if ~strcmp(iVerbosity,'movie') 
      return; 
    end;
  end;
    
  %Draw models on side of image
  t.szt = size(iMdl.T0); 
  t.szi = size(fI);
  tI = fI;
  tI(1:t.szt(1),t.szi(2)+1:t.szi(2)+t.szt(2)) = iMdl.T0;
  tI(t.szt(1)+1:t.szt(1)*2,t.szi(2)+1:t.szi(2)+t.szt(2)) = iMdl.TN;
  
  figure(2); 
  clf;
  
  image( tI );
  colormap(gray(256));

  set(gcf,'Position',[5 5 fliplr(size(tI))*2]);
  set(gca,'Position',[0 0 1 1]);
  
  %Previous position
  if size(iHist,2)>1
    fX = iHist(1:6,end-1);
    [fRect,y,x] = getRect(size(iMdl.T0),fX);
    patch( x,y, [0 1 0],'FaceAlpha',0.1,'EdgeColor',[0 1 1]);
  end
  
  %Intermediate positions  
  ct = size(iOptHist.x,1);
  for c1 = 1:ct,
    fX = iOptHist.x(c1,1:6);
    [fRect,y,x] = getRect(size(iMdl.T0),fX);
    th = patch( x,y,[0 1 0],'FaceAlpha',0.0,'EdgeColor',[0 c1/ct (ct-c1)/ct]);
    if c1==ct, set(th,'FaceAlpha',0.1); end
  end
  
  %Current position
  fX = iHist(1:6,end);
  hold on;
  plot(fX(6),fX(5),'g+');
  %figure(3); clf;
  %if size(iOptHist.x,1) > 1, plot([1:6]*10,iOptHist.x,'r-'); end
  %if size(iOptHist.x,1) > 1,
    %t.y = iOptHist.y;
    t.y2 = iHist(7,:);
    if t.y2(1)<1e-4, t.ylo = min(t.y2);
    else t.ylo = 0; end;
    t.yhi = max(t.y2);
    t.yrng = t.yhi - t.ylo;
    if length(t.y2) < size(fI,2)
      t.x2 = [0:length(t.y2)-1]; 
    else
      t.x2 = [0:length(t.y2)-1] / length(t.y2) * size(fI,2);
    end
    
    t.y1 = iOptHist.y;
    t.xlo = size(fI,2)/20;
    t.ihi = size(fI,1);
    if t.yrng==0, t.yrng=1; end
    t.x1 = [0:length(t.y1)-1]*5;% / length(t.y1) * max(t.x2);
    
    plot(t.x1,t.ihi-(t.y1-t.ylo)/t.yrng*t.ihi/10,'y-',...
         'LineWidth',1.5);
    plot(t.x2,t.ihi-(t.y2-t.ylo)/t.yrng*t.ihi/10,'g-',...
         'LineWidth',1.5);
  %end
   
  tTime = floor(totalTime*100)/100;
  text( 10,10,['Frm ' num2str(size(iHist,2)) ' - ' num2str(tTime) 's'],...
        'BackgroundColor',[1 1 1]);
  text( 10,20, ['LM iter ' num2str(size(iOptHist.x,1))],...
        'BackgroundColor',[1 1 1]);
  if exist('iParam','var')
    text( size(tI,2),size(tI,1), ...
          [iParam{6}{1} ' , ' iParam{3} '(' num2str(iParam{4}) ')'],...
          'BackgroundColor',[1 1 1], ...
          'VerticalAlignment','bottom', 'HorizontalAlignment','right');  
  end
  pause(0.01);
  
  fFrm = size(iHist,2);
  if strcmp(iVerbosity,'movie')
    if fFrm<10, fPad = '000';
    elseif fFrm<100, fPad = '00';
    elseif fFrm<1000, fPad = '0';
    else fPad = '';
    end
    set(gcf, 'PaperPositionMode', 'auto');
    print('-dpng',['vid/' fPad num2str(fFrm) '.png']);
  end
  
  tic;

%---------------------------------------------------------------------
function [oRect oX oY] = getRect(iSz,iWarp)
% GETRECT - Return a warped rectangle
%   

  iSz = (iSz-1)/2;
  x = [-1  1  1 -1];
  y = [-1 -1  1  1];
  z = [ 1  1  1  1];
  fRect = [x*iSz(1); y*iSz(2); z];
  fRot = reshape(iWarp(:)+[1 0 0 1 0 0]',[2 3]);
  
  oRect = fRot*fRect;
  oX = oRect(1,:);
  oY = oRect(2,:);